'use strict';

const fs = require('node:fs');
const node_path = require('node:path');
const process = require('node:process');
const shared = require('@tailwindcss-mangle/shared');
const config = require('@tailwindcss-mangle/config');
const fastSort = require('fast-sort');
const postcss = require('postcss');
const parser = require('postcss-selector-parser');
const htmlparser2 = require('htmlparser2');
const MagicString = require('magic-string');
const escape = require('@ast-core/escape');
const _babelTraverse = require('@babel/traverse');
const parser$1 = require('@babel/parser');

function _interopDefaultCompat (e) { return e && typeof e === 'object' && 'default' in e ? e.default : e; }

const fs__default = /*#__PURE__*/_interopDefaultCompat(fs);
const process__default = /*#__PURE__*/_interopDefaultCompat(process);
const postcss__default = /*#__PURE__*/_interopDefaultCompat(postcss);
const parser__default = /*#__PURE__*/_interopDefaultCompat(parser);
const MagicString__default = /*#__PURE__*/_interopDefaultCompat(MagicString);
const _babelTraverse__default = /*#__PURE__*/_interopDefaultCompat(_babelTraverse);

function isPlainObject(value) {
  if (value === null || typeof value !== "object") {
    return false;
  }
  const prototype = Object.getPrototypeOf(value);
  if (prototype !== null && prototype !== Object.prototype && Object.getPrototypeOf(prototype) !== null) {
    return false;
  }
  if (Symbol.iterator in value) {
    return false;
  }
  if (Symbol.toStringTag in value) {
    return Object.prototype.toString.call(value) === "[object Module]";
  }
  return true;
}

function _defu(baseObject, defaults, namespace = ".", merger) {
  if (!isPlainObject(defaults)) {
    return _defu(baseObject, {}, namespace, merger);
  }
  const object = Object.assign({}, defaults);
  for (const key in baseObject) {
    if (key === "__proto__" || key === "constructor") {
      continue;
    }
    const value = baseObject[key];
    if (value === null || value === void 0) {
      continue;
    }
    if (merger && merger(object, key, value, namespace)) {
      continue;
    }
    if (Array.isArray(value) && Array.isArray(object[key])) {
      object[key] = [...value, ...object[key]];
    } else if (isPlainObject(value) && isPlainObject(object[key])) {
      object[key] = _defu(
        value,
        object[key],
        (namespace ? `${namespace}.` : "") + key.toString(),
        merger
      );
    } else {
      object[key] = value;
    }
  }
  return object;
}
function createDefu(merger) {
  return (...arguments_) => (
    // eslint-disable-next-line unicorn/no-array-reduce
    arguments_.reduce((p, c) => _defu(p, c, "", merger), {})
  );
}
const defu = createDefu();

function escapeStringRegexp(str) {
  if (typeof str !== "string") {
    throw new TypeError("Expected a string");
  }
  return str.replaceAll(/[$()*+.?[\\\]^{|}]/g, "\\$&").replaceAll("-", "\\x2d");
}

var __defProp = Object.defineProperty;
var __defNormalProp = (obj, key, value) => key in obj ? __defProp(obj, key, { enumerable: true, configurable: true, writable: true, value }) : obj[key] = value;
var __publicField = (obj, key, value) => {
  __defNormalProp(obj, typeof key !== "symbol" ? key + "" : key, value);
  return value;
};
class Context {
  constructor() {
    __publicField(this, "options");
    __publicField(this, "replaceMap");
    __publicField(this, "classSet");
    __publicField(this, "classGenerator");
    __publicField(this, "preserveFunctionSet");
    __publicField(this, "preserveClassNamesSet");
    __publicField(this, "preserveFunctionRegexs");
    this.options = {};
    this.classSet = /* @__PURE__ */ new Set();
    this.replaceMap = /* @__PURE__ */ new Map();
    this.classGenerator = new shared.ClassGenerator();
    this.preserveFunctionSet = /* @__PURE__ */ new Set();
    this.preserveClassNamesSet = /* @__PURE__ */ new Set();
    this.preserveFunctionRegexs = [];
  }
  isPreserveClass(className) {
    return this.preserveClassNamesSet.has(className);
  }
  addPreserveClass(className) {
    return this.preserveClassNamesSet.add(className);
  }
  isPreserveFunction(calleeName) {
    return this.preserveFunctionSet.has(calleeName);
  }
  mergeOptions(...opts) {
    this.options = defu(this.options, ...opts);
    this.classGenerator = new shared.ClassGenerator(this.options.classGenerator);
    this.preserveFunctionSet = new Set(this.options?.preserveFunction ?? []);
    this.preserveFunctionRegexs = [...this.preserveFunctionSet.values()].map((x) => {
      return new RegExp(`${escapeStringRegexp(x)}\\(([^)]*)\\)`, "g");
    });
  }
  currentMangleClassFilter(className) {
    return (this.options.mangleClassFilter ?? shared.defaultMangleClassFilter)(className);
  }
  getClassSet() {
    return this.classSet;
  }
  getReplaceMap() {
    return this.replaceMap;
  }
  addToUsedBy(key, file) {
    if (!file) {
      return;
    }
    const hit = this.classGenerator.newClassMap[key];
    if (hit) {
      hit.usedBy.add(file);
    }
  }
  loadClassSet(classList) {
    const list = fastSort.sort(classList).desc((c) => c.length);
    for (const className of list) {
      if (this.currentMangleClassFilter(className)) {
        this.classSet.add(className);
      }
    }
  }
  async initConfig(opts = {}) {
    const { cwd, classList: _classList, mangleOptions } = opts;
    const { config: config$1, cwd: configCwd } = await config.getConfig(cwd);
    if (mangleOptions?.classMapOutput === true) {
      mangleOptions.classMapOutput = config$1.mangle?.classMapOutput;
      if (typeof mangleOptions.classMapOutput === "object") {
        mangleOptions.classMapOutput.enable = true;
      }
    }
    this.mergeOptions(mangleOptions, config$1?.mangle);
    if (_classList) {
      this.loadClassSet(_classList);
    } else {
      let jsonPath = this.options.classListPath ?? node_path.resolve(process__default.cwd(), config$1?.patch?.output?.filename);
      if (!node_path.isAbsolute(jsonPath)) {
        jsonPath = node_path.resolve(configCwd ?? process__default.cwd(), jsonPath);
      }
      if (jsonPath && fs__default.existsSync(jsonPath)) {
        const rawClassList = fs__default.readFileSync(jsonPath, "utf8");
        const list = JSON.parse(rawClassList);
        this.loadClassSet(list);
      }
    }
    for (const cls of this.classSet) {
      this.classGenerator.generateClassName(cls);
    }
    for (const x of Object.entries(this.classGenerator.newClassMap)) {
      this.replaceMap.set(x[0], x[1].name);
    }
    return config$1;
  }
  async dump() {
    try {
      const arr = Object.entries(this.classGenerator.newClassMap).map((x) => {
        return {
          before: x[0],
          after: x[1].name,
          usedBy: Array.from(x[1].usedBy)
        };
      });
      if (typeof this.options.classMapOutput === "function") {
        await this.options.classMapOutput(arr);
      } else if (typeof this.options.classMapOutput === "object" && this.options.classMapOutput.enable && this.options.classMapOutput.filename) {
        fs__default.mkdirSync(node_path.dirname(this.options.classMapOutput.filename), { recursive: true });
        fs__default.writeFileSync(this.options.classMapOutput.filename, JSON.stringify(arr, null, 2));
      }
    } catch (error) {
      console.error(`[tailwindcss-mangle]: ${error}`);
    }
  }
}

const postcssPlugin = "postcss-mangle-tailwindcss-plugin";
function isVueScoped(s) {
  if (s.parent) {
    const index = s.parent.nodes.indexOf(s);
    if (index > -1) {
      const nextNode = s.parent.nodes[index + 1];
      if (nextNode && nextNode.type === "attribute" && nextNode.attribute.includes("data-v-")) {
        return true;
      }
    }
  }
  return false;
}
const transformSelectorPostcssPlugin = function(options) {
  const { ignoreVueScoped, ctx, id } = defu(options, {
    ignoreVueScoped: true
  });
  const replaceMap = ctx.replaceMap;
  return {
    postcssPlugin,
    Once(root) {
      root.walkRules((rule) => {
        parser__default((selectors) => {
          selectors.walkClasses((s) => {
            if (s.value && replaceMap && replaceMap.has(s.value)) {
              if (ignoreVueScoped && isVueScoped(s)) {
                return;
              }
              const v = replaceMap.get(s.value);
              if (v) {
                if (ctx.isPreserveClass(s.value)) {
                  rule.cloneBefore();
                }
                s.value = v;
              }
            }
          });
        }).transformSync(rule, {
          lossless: false,
          updateSelector: true
        });
      });
    }
  };
};
transformSelectorPostcssPlugin.postcss = true;

async function cssHandler(rawSource, options) {
  const acceptedPlugins = [transformSelectorPostcssPlugin(options)];
  const { id } = options;
  const { css: code, map } = await postcss__default(acceptedPlugins).process(rawSource, {
    from: id,
    to: id
  });
  return {
    code,
    // @ts-ignore
    map
  };
}

function htmlHandler(raw, options) {
  const { ctx, id } = options;
  const { replaceMap, classGenerator } = ctx;
  const ms = typeof raw === "string" ? new MagicString__default(raw) : raw;
  const parser = new htmlparser2.Parser({
    onattribute(name, value) {
      if (name === "class") {
        let needUpdate = false;
        const arr = shared.splitCode(value, {
          splitQuote: false
        });
        let rawValue = value;
        for (const v of arr) {
          if (replaceMap.has(v)) {
            const gen = classGenerator.generateClassName(v);
            rawValue = rawValue.replace(shared.makeRegex(v), gen.name);
            ctx.addToUsedBy(v, id);
            needUpdate = true;
          }
        }
        needUpdate && ms.update(parser.startIndex + name.length + 2, parser.endIndex - 1, rawValue);
      }
    }
  });
  parser.write(ms.original);
  parser.end();
  return {
    code: ms.toString()
  };
}

function _interopDefaultCompat$1(e) {
  return e && typeof e === "object" && "default" in e ? e.default : e;
}
const traverse = _interopDefaultCompat$1(_babelTraverse__default);

function handleValue(raw, node, options, ms, offset, escape$1) {
  const { ctx, splitQuote = true, id } = options;
  const { replaceMap, classGenerator: clsGen } = ctx;
  const array = shared.splitCode(raw, {
    splitQuote
  });
  let rawString = raw;
  let needUpdate = false;
  for (const v of array) {
    if (replaceMap.has(v)) {
      let ignoreFlag = false;
      if (Array.isArray(node.leadingComments)) {
        ignoreFlag = node.leadingComments.findIndex((x) => x.value.includes("tw-mangle") && x.value.includes("ignore")) > -1;
      }
      if (!ignoreFlag) {
        const gen = clsGen.generateClassName(v);
        rawString = rawString.replace(shared.makeRegex(v), gen.name);
        ctx.addToUsedBy(v, id);
        needUpdate = true;
      }
    }
  }
  if (needUpdate && typeof node.start === "number" && typeof node.end === "number") {
    const start = node.start + offset;
    const end = node.end - offset;
    if (start < end && raw !== rawString) {
      ms.update(start, end, escape$1 ? escape.jsStringEscape(rawString) : rawString);
    }
  }
  return rawString;
}
function jsHandler(rawSource, options) {
  const ms = typeof rawSource === "string" ? new MagicString__default(rawSource) : rawSource;
  let ast;
  try {
    ast = parser$1.parse(ms.original, {
      sourceType: "unambiguous"
    });
  } catch (error) {
    return {
      code: ms.original
    };
  }
  const { ctx } = options;
  traverse(ast, {
    StringLiteral: {
      enter(p) {
        const n = p.node;
        handleValue(n.value, n, options, ms, 1, true);
      }
    },
    TemplateElement: {
      enter(p) {
        const n = p.node;
        handleValue(n.value.raw, n, options, ms, 0, false);
      }
    },
    CallExpression: {
      enter(p) {
        const callee = p.get("callee");
        if (callee.isIdentifier() && ctx.isPreserveFunction(callee.node.name)) {
          p.traverse({
            StringLiteral: {
              enter(path) {
                const node = path.node;
                const value = node.value;
                const arr = fastSort.sort(shared.splitCode(value)).desc((x) => x.length);
                for (const str of arr) {
                  if (ctx.replaceMap.has(str)) {
                    ctx.addPreserveClass(str);
                  }
                }
              }
            },
            TemplateElement: {
              enter(path) {
                const node = path.node;
                const value = node.value.raw;
                const arr = fastSort.sort(shared.splitCode(value)).desc((x) => x.length);
                for (const str of arr) {
                  if (ctx.replaceMap.has(str)) {
                    ctx.addPreserveClass(str);
                  }
                }
              }
            }
          });
        }
      }
    }
  });
  return {
    code: ms.toString(),
    get map() {
      return ms.generateMap();
    }
  };
}

exports.ClassGenerator = shared.ClassGenerator;
exports.Context = Context;
exports.cssHandler = cssHandler;
exports.handleValue = handleValue;
exports.htmlHandler = htmlHandler;
exports.jsHandler = jsHandler;
